// Copyright 2011-2012 Daniele Panozzo, panozzo@inf.ethz.ch. All rights reserved.
// The code in this file cannot be used for commercial
// applications without prior written permission from Daniele Panozzo

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <GL/glew.h>

#include "visualizer.h"
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <AntTweakBar.h>

#include "state.h"

#include <QtCore/QDir>

#include <cmath>
#include <iostream>


//#include <Eigen/Sparse>
#define WindowW 1280
#define WindowH 720

Visualizer *visualizer;
int tw_my_var = 0;
bool movierunning;
int moviecount;

void display ()
{
    visualizer->paintGL ();
    TwDraw ();
    glutSwapBuffers ();
}

void reshape (int w, int h)
{
    visualizer->resizeGL (w, h);
    TwWindowSize (w, h);
    glutPostRedisplay ();
}

void mouse_click (int button, int state, int x, int y)
{
    if (TwEventMouseButtonGLUT (button, state, x, y))
    {
		glutPostRedisplay ();
        return;
    }
    
	visualizer->mousePress(button, state, x, y);
	glutPostRedisplay ();
}

void mouse_move (int x, int y)
{
    if (TwEventMouseMotionGLUT (x, y))
    {
		glutPostRedisplay ();
        return;
    }
	
    visualizer->mouseMove (x, y);
}

void saveScreenshot(int i)
{
    
    string temp = State::Instance()->filename;
    stringstream ss;
    ss << temp << "_movie";
    QDir().mkdir(ss.str().c_str());
    
    ss << "/";
    ss.width(5);
    ss.fill('0');
    ss << right << moviecount++;
    
    visualizer->saveSnapshot(ss.str().c_str());    
    
    if (movierunning)
        glutTimerFunc(1000/30, saveScreenshot, 1);
}

void keyboard (unsigned char k, int x, int y)
{
    if (k == 'o')
    {
        movierunning = true;
        moviecount = 0;
        glutTimerFunc(1000/30, saveScreenshot, 1);
    }
    if (k == 'p')
    {
        movierunning = false;
    }    
    
    if (k == 'g')
    {
        visualizer->meshEnabled  = !visualizer->meshEnabled;
        visualizer->meshEnabled2 = !visualizer->meshEnabled2;
    }
    
    if (k == 'f')
    {
        State::Instance()->solveUniform();
    }
    
    if (k == 'd')
    {
        State::Instance()->solve();
    }
    
    TwEventKeyboardGLUT (k, x, y);
    visualizer->keyPress (k);
	glutPostRedisplay();
}

void special (int k, int x, int y)
{
    TwEventSpecialGLUT (k, x, y);
	glutPostRedisplay();
}

void TW_CALL SetW(const void *value, void *clientData)
{ 
    State::Instance()->currentSize[0] =*(const double *)value;
    State::Instance()->moveCrop();
    State::Instance()->solve();
}

void TW_CALL GetW(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->currentSize[0];
}

void TW_CALL SetH(const void *value, void *clientData)
{ 
    State::Instance()->currentSize[1] =*(const double *)value;
    State::Instance()->moveCrop();
    State::Instance()->solve();
}

void TW_CALL GetH(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->currentSize[1];
}

void TW_CALL SetMinW(const void *value, void *clientData)
{ 
    State::Instance()->cropBL[0] =*(const double *)value;
    if (State::Instance()->cropBL[0] >= State::Instance()->cropTR[0])
        State::Instance()->cropBL[0] = State::Instance()->cropTR[0] - 5;
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetMinW(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->cropBL[0];
}
void TW_CALL SetMinH(const void *value, void *clientData)
{ 
    State::Instance()->cropBL[1] =*(const double *)value;
    if (State::Instance()->cropBL[1] >= State::Instance()->cropTR[1])
        State::Instance()->cropBL[1] = State::Instance()->cropTR[1] - 5;
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetMinH(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->cropBL[1];
}
void TW_CALL SetMaxW(const void *value, void *clientData)
{ 
    State::Instance()->cropTR[0] =*(const double *)value;
    if (State::Instance()->cropTR[0] <= State::Instance()->cropBL[0])
        State::Instance()->cropTR[0] = State::Instance()->cropBL[0] + 5;
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetMaxW(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->cropTR[0];
}
void TW_CALL SetMaxH(const void *value, void *clientData)
{ 
    State::Instance()->cropTR[1] =*(const double *)value;
    if (State::Instance()->cropTR[1] <= State::Instance()->cropBL[1])
        State::Instance()->cropTR[1] = State::Instance()->cropBL[1] + 5;
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetMaxH(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->cropTR[1];
}

// --- Two knots crop!
void TW_CALL GetCropL(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->crop2P[0];
}
void TW_CALL SetCropL(const void *value, void *clientData)
{ 
    State::Instance()->crop2P[0] =*(const double *)value;
    State::Instance()->moveCrop();
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetCropH(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->crop2P[1];
}

void TW_CALL SetCropH(const void *value, void *clientData)
{ 
    State::Instance()->crop2P[1] =*(const double *)value;
    State::Instance()->moveCrop();
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

void TW_CALL GetCrop(void *value, void *clientData)
{ 
    *(double *)value = (State::Instance()->crop2P[0] + State::Instance()->crop2P[1])/2.0;
}

void TW_CALL SetCrop(const void *value, void *clientData)
{ 
    State::Instance()->crop2P[0] =*(const double *)value;
    State::Instance()->crop2P[1] =*(const double *)value;
    State::Instance()->moveCrop();
    State::Instance()->updateCrop();
    State::Instance()->solve();
}

// ---


void TW_CALL SetLW(const void *value, void *clientData)
{ 
    State* st = State::Instance(); 
    st->LWfactor = *(const double *)value;
    State::Instance()->solve();
}

void TW_CALL GetLW(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->LWfactor;
}

void TW_CALL SetLH(const void *value, void *clientData)
{ 
    State* st = State::Instance(); 
    st->LHfactor = *(const double *)value;
    State::Instance()->solve();
}

void TW_CALL GetLH(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->LHfactor;
}

void TW_CALL SetLapW(const void *value, void *clientData)
{ 
    State::Instance()->laplacianWeight = *(const double *)value;
    State::Instance()->solve();
}

void TW_CALL GetLapW(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->laplacianWeight;
}

void TW_CALL SetLapE(const void *value, void *clientData)
{ 
    State::Instance()->addLaplacianTermsEnabled = *(const bool *)value;
    State::Instance()->solve();
}

void TW_CALL GetLapE(void *value, void *clientData)
{ 
    *(bool *)value = State::Instance()->addLaplacianTermsEnabled;
}

void TW_CALL SetAAAlpha(const void *value, void *clientData)
{ 
    State::Instance()->HYBRID_alpha = *(const double *)value;
    State::Instance()->solve();
}

void TW_CALL GetAAAlpha(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->HYBRID_alpha;
}

void TW_CALL SetEnergy(const void *value, void *clientData)
{ 
    State* st = State::Instance(); 
    st->energyType =*(const State::EnergyType *)value;
    
    //    if (st->energyType == State::ENERGY_ARAP)
    //    st->laplacianWeight = 0.005;
    
    //    if (st->energyType == State::ENERGY_ASAP)
    //    st->laplacianWeight = 0.5;
    
    State::Instance()->solve(true);
}

void TW_CALL GetEnergy(void *value, void *clientData)
{ 
    *(State::EnergyType *)value = State::Instance()->energyType;
}

void TW_CALL SetBrightness(const void *value, void *clientData)
{ 
    State::Instance()->brightness = *(const double *)value;
    State::Instance()->updateLoaded();
    State::Instance()->updateSaliency();
    State::Instance()->solve();
}

void TW_CALL GetBrightness(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->brightness;
}

void TW_CALL SetContrast(const void *value, void *clientData)
{ 
    State::Instance()->contrast = *(const double *)value;
    State::Instance()->updateLoaded();
    State::Instance()->updateSaliency();
    State::Instance()->solve();
}

void TW_CALL GetContrast(void *value, void *clientData)
{ 
    *(double *)value = State::Instance()->contrast;
}

void TW_CALL Solve(void * clientData)
{ 
    State::Instance()->solve(true);
}

void TW_CALL SolveUniform(void * clientData)
{ 
    State::Instance()->solveUniform();
}

void TW_CALL Save(void * clientData)
{ 
    string temp = State::Instance()->filename;
    State::Instance()->solveUniform();
    visualizer->renderToFile((temp + string(".uniform.png")).c_str());
    State::Instance()->solve(true);
    visualizer->renderToFile((temp + string(".resized.png")).c_str());
}

void TW_CALL SaveSaliency(void * clientData)
{ 
    string temp = State::Instance()->filename;
    State::Instance()->saveSaliency((temp.substr(0,temp.size()-4) + string("_saliency_manual.png")).c_str());
}

void TW_CALL SaveAll(void * clientData)
{ 
    Save(clientData);
    SaveSaliency(clientData);
    string temp = State::Instance()->filename;
    State::Instance()->saveParams((temp + string(".params.txt")).c_str());
    State::Instance()->saveMap((temp + string(".map.txt")).c_str());

}

void TW_CALL ResetView(void * clientData)
{ 
	Visualizer* v = (Visualizer*) clientData;
	v->resetView();
}

#ifdef VARGRID
void TW_CALL ChangeGrid(void * clientData)
{ 
	Visualizer* v = (Visualizer*) clientData;
    State::Instance()->initGrid();
    State::Instance()->initSaliency(false);
}
#endif

void TW_CALL ClearSaliency(void * clientData)
{ 
    State::Instance()->clearSaliency();
}

void TW_CALL Gradient(void * clientData)
{ 
    State::Instance()->computeGradient(State::Instance()->filename.c_str());
    State::Instance()->solve(true);
}

void TW_CALL Restore(void * clientData)
{ 
    
}

void TW_CALL SaveObj(void * clientData)
{ 
    
}

int main(int argc, char *argv[])
{
    glutInit (&argc, argv);
    glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
    glutInitWindowSize (WindowW, WindowH);
    glutCreateWindow ("Robust Interactive Image Retargeting via Axis-Aligned Deformation - Copyright 2011 Daniele Panozzo, Ofir Weber, Olga Sorkine");
	
    char* image;
    movierunning = false;
    
    double factor = 1.0/2.0;
    
    if (argc >= 2 && string(argv[1]) == string("-batch"))
    {
        // batch mode
        
        if (argc != 8 && argc != 9)
        {
            cout << "Usage: ./SIR -batch src dst w h lw [ASAP ARAP UNIFORM] [sal]";
            exit(1);
        }
        
        image = argv[2];
        string ori(argv[2]);
        string saveTo(argv[3]);
        
        double scalefactorx = atof(argv[4]);
        double scalefactory = atof(argv[5]);
        
        double alpha = atof(argv[6]);
        
        string resizetype(argv[7]);
        
        visualizer = new Visualizer (WindowW, WindowH);
        
        State* st = State::Instance();
        st->exportMode = true;
        
        st->loadImage(image);
        
        visualizer->initializeGL();
        visualizer->initMesh();
        
        if (argc != 9)
            st->computeGradient(st->filename.c_str());
        else
            st->loadSaliency(argv[8]);
        
        st->currentSize[0] = st->originalSize[0] * scalefactorx;
        st->currentSize[1] = st->originalSize[1] * scalefactory;
        
        State::Instance()->laplacianWeight = alpha;
        
        if (resizetype == string("ASAP"))
        {
            st->energyType = State::ENERGY_ASAP;
            State::Instance()->solve();
            visualizer->renderToFile(saveTo.c_str());
        }
        else if (resizetype == string("ARAP"))
        {
            st->energyType = State::ENERGY_ARAP;
            State::Instance()->solve();
            visualizer->renderToFile(saveTo.c_str());
        }
        else if (resizetype == string("UNIFORM"))
        {
            State::Instance()->solveUniform();
            visualizer->renderToFile(saveTo.c_str());
        }
        else if (resizetype == string("SALIENCY"))
        {
            State::Instance()->solveUniform();
            st->saveSaliency(saveTo.c_str());
        }
        
        exit(0);
        
    }
    
    if (argc != 2 && argc !=3 && argc != 4)
        image = "child.png";
    else
        image = argv[1];
    
    TwInit (TW_OPENGL, NULL);
    TwWindowSize (WindowW, WindowH);
    
    visualizer = new Visualizer (WindowW, WindowH);
    
	State* st = State::Instance();
    
    st->loadImage(image);
    
    visualizer->initializeGL();
    visualizer->initMesh();
    
    if (argc == 3 || argc == 4) // saliency provided
        st->loadSaliency(argv[2]);
    else
        st->computeGradient(image);
	
    glutDisplayFunc (display);
    glutReshapeFunc (reshape);
    glutMouseFunc (mouse_click);
    glutMotionFunc (mouse_move);
    glutPassiveMotionFunc (mouse_move);
    glutKeyboardFunc (keyboard);
    glutSpecialFunc (special);
    TwGLUTModifiersFunc (glutGetModifiers);
	
    TwBar *cBar;
    cBar = TwNewBar ("IR");
    TwDefine("IR position='1050 100'");
    TwDefine("IR size='200 500'");
    TwDefine("IR valueswidth=75");
    TwDefine("GLOBAL help='Robust Interactive Image Retargeting via Axis-Aligned Deformation. Daniele Panozzo, Ofir Weber, Olga Sorkine'");
    //TwDefine("Visualizer color='192 255 192' text=dark ");
	
#define RECORDMODE 0
    
    if (RECORDMODE)
    {
        
        TwEnumVal energyTypeEV[ENERGY_TYPE_NUM] = 
        { 
            {State::ENERGY_ASAP, "ASAP"},
            {State::ENERGY_ARAP, "ARAP"},
            {State::ENERGY_HYBRID, "HYBRID"}
        };
        
        TwType solverTypeT = TwDefineEnum("energyModeType", energyTypeEV, ENERGY_TYPE_NUM);
        TwAddVarCB(cBar, "Energy", solverTypeT, &SetEnergy, &GetEnergy, NULL, " keyDecr='.' help='Energy to minimize.'");
        TwAddVarCB(cBar, "Hybrid weight", TW_TYPE_DOUBLE, &SetAAAlpha, &GetAAAlpha, NULL, "min=0 max=1 step =0.01");
        
        
        TwAddButton(cBar, "Clear Saliency", ClearSaliency, visualizer, "");
        
        TwAddVarCB(cBar, "Brightness ", TW_TYPE_DOUBLE, &SetBrightness, &GetBrightness, NULL, "min=0 max=1 step=0.001");
        TwAddVarCB(cBar, "Contrast   ", TW_TYPE_DOUBLE, &SetContrast, &GetContrast, NULL, "min=-1 max=1 step=0.001");
        
        //st->useExtGrid = false;
        TwAddButton(cBar, "Save", SaveAll, visualizer, "");
        
        TwAddVarCB(cBar, "W", TW_TYPE_DOUBLE, &SetW, &GetW, NULL, "min = 10");
        TwAddVarCB(cBar, "H", TW_TYPE_DOUBLE, &SetH, &GetH, NULL, "min = 10");
        TwAddVarCB(cBar, "LH", TW_TYPE_DOUBLE, &SetLW, &GetLW, NULL, "min=0 max=0.5 step =0.001");
        TwAddVarCB(cBar, "LW", TW_TYPE_DOUBLE, &SetLH, &GetLH, NULL, "min=0 max=0.5 step =0.001");
        
        TwAddVarCB(cBar, "Laplacian Weight", TW_TYPE_DOUBLE, &SetLapW, &GetLapW, NULL, "min=0 max=1000 step =0.01");
        
        
        TwAddVarRW(cBar, "Brush Size", TW_TYPE_DOUBLE, &st->brushRadius, "min=0 max=1000 step = 20 keyDecr='k' keyIncr='l'");
        
    }
    else
    {
        TwAddVarRW(cBar, "(L.) Grid", TW_TYPE_BOOLCPP, &visualizer->meshEnabled, "help='Show the grid on the left'");
        TwAddVarRW(cBar, "(L.) Image", TW_TYPE_BOOLCPP, &visualizer->textureEnabled, "help='Show the image on the left'");
        TwAddVarRW(cBar, "(L.) Saliency", TW_TYPE_BOOLCPP, &visualizer->saliencyEnabled, "help='Show the saliency map on the left'");
        
        TwAddVarRW(cBar, "(R.) Grid", TW_TYPE_BOOLCPP, &visualizer->meshEnabled2, "help='Show the grid on the right'");
        //TwAddVarRW(cBar, "(R.) Image", TW_TYPE_BOOLCPP, &visualizer->textureEnabled2, "");
        //TwAddVarRW(cBar, "(R.) Saliency", TW_TYPE_BOOLCPP, &visualizer->saliencyEnabled2, "");
        
        TwEnumVal energyTypeEV[ENERGY_TYPE_NUM] = 
        { 
            {State::ENERGY_ASAP, "ASAP"},
            {State::ENERGY_ARAP, "ARAP"},
            {State::ENERGY_HYBRID, "HYBRID"}
        };
        
        TwType solverTypeT = TwDefineEnum("energyModeType", energyTypeEV, ENERGY_TYPE_NUM);
        TwAddVarCB(cBar, "Energy", solverTypeT, &SetEnergy, &GetEnergy, NULL, " keyDecr='.' help='Energy to minimize.'");
        TwAddVarCB(cBar, "Hybrid weight", TW_TYPE_DOUBLE, &SetAAAlpha, &GetAAAlpha, NULL, "min=0 max=1 step =0.001 help='If the energy is hybrid, this parameter interpolates between ASAP and ARAP (0: ASAP, 1:ARAP)'");
        
        
        TwEnumVal salyencyTypeEV[SALIENCY_TYPE_NUM] = 
        { 
            {State::SAL_USER, "USER"},
            {State::SAL_LOADED, "LOADED"},
            {State::SAL_BOTH, "BOTH"},
        };
        
        //    TwAddVarCB(cBar, "Alpha (ASAPARAP)", TW_TYPE_DOUBLE, &SetAAAlpha, &GetAAAlpha, NULL, "min=0 max=1 step =0.001");
        
        
        TwType saliencyTypeT = TwDefineEnum("saliencyModeType", salyencyTypeEV, SALIENCY_TYPE_NUM);
        TwAddVarRW(cBar, "Saliency", saliencyTypeT, &st->saliencyType, " keyDecr=',' help='Saliency map to use (USER: painted saliency, LOADED: map loaded from a file, BOTH: sum of USER and LOADED)'");
        
        TwAddSeparator (cBar, 0, 0);
        //	TwAddButton(cBar, "Reset View", ResetView, visualizer, "");
        TwAddButton(cBar, "Clear Saliency", ClearSaliency, visualizer, "");
        //TwAddButton(cBar, "Compute Gradient", Gradient, visualizer, "");
        TwAddVarCB(cBar, "Brightness ", TW_TYPE_DOUBLE, &SetBrightness, &GetBrightness, NULL, "min=0 max=1 step=0.001");
        TwAddVarCB(cBar, "Contrast   ", TW_TYPE_DOUBLE, &SetContrast, &GetContrast, NULL, "min=-1 max=1 step=0.001");
        
        // View Options
        TwAddVarRW(cBar, "Gridcolor", TW_TYPE_COLOR3F, &visualizer->gridColor, " colormode=rgb");
        //TwAddSeparator (cBar, 0, 0);
        
        // Buttons
        // ...
        TwAddVarRW(cBar, "Extended Grid", TW_TYPE_BOOLCPP, &st->useExtGrid, "help='Use spline interpolation (useful only for large images)'");
        
        TwAddButton(cBar, "Save", SaveAll, visualizer, "help='Save in PNG format the painted saliency and the retargeted image'");
        
        TwAddSeparator (cBar, 0, 0);
        
        TwAddVarCB(cBar, "W", TW_TYPE_DOUBLE, &SetW, &GetW, NULL, "min = 10 help='Change the width of the retargeted image'");
        TwAddVarCB(cBar, "H", TW_TYPE_DOUBLE, &SetH, &GetH, NULL, "min = 10 help='Change the height of the retargeted image'");
        TwAddVarCB(cBar, "Min. H", TW_TYPE_DOUBLE, &SetLW, &GetLW, NULL, "min=0 max=0.5 step =0.001 help='Change the minimal width of every column in the retargeted grid'");
        TwAddVarCB(cBar, "Min. W", TW_TYPE_DOUBLE, &SetLH, &GetLH, NULL, "min=0 max=0.5 step =0.001 help='Change the minimal height of every row in the retargeted grid'");
        
        TwAddSeparator (cBar, 0, 0);
        TwAddVarCB(cBar, "Laplacian E. ", TW_TYPE_BOOLCPP, &SetLapE, &GetLapE, NULL, "help='Add to the energy the laplacian regularization term'");
        TwAddVarCB(cBar, "Laplacian Weight", TW_TYPE_DOUBLE, &SetLapW, &GetLapW, NULL, "min=0 max=1000 step =0.01 help='Weight of the laplacian regularization term'");
        
        TwAddSeparator (cBar, 0, 0);
        
        std::stringstream out ; out  << "min = 0 max = " << st->originalSize[0]; string tempW = out .str();
        std::stringstream out2; out2 << "min = 0 max = " << st->originalSize[1]; string tempH = out2.str();
        
        TwAddVarCB(cBar, "Crop Low" , TW_TYPE_DOUBLE, &SetCropL, &GetCropL, NULL, "min=0 max=0.5 step =0.01 help='Crop the bottom or left part of the image, depending on the aspect ratio of the retargeted image'");
        TwAddVarCB(cBar, "Crop High" , TW_TYPE_DOUBLE, &SetCropH, &GetCropH, NULL, "min=0 max=0.5 step =0.01 help='Crop the top or right part of the image, depending on the aspect ratio of the retargeted image'");

        TwAddVarCB(cBar, "Crop" , TW_TYPE_DOUBLE, &SetCrop, &GetCrop, NULL, "min=0 max=0.5 step =0.01 help='Crop the image horizontally or vertically, depending on the aspect ratio of the retargeted image'");

        TwAddSeparator (cBar, 0, 0);
        
        TwAddVarRW(cBar, "Brush Size", TW_TYPE_DOUBLE, &st->brushRadius, "min=0 max=1000 step = 20 keyDecr='k' keyIncr='l' help='Size of the brush used to paint the saliency map, keep ctrl pressed to paint and alt pressed to erase'");
        
#ifdef VARGRID
        TwAddSeparator (cBar, 0, 0);
        TwAddVarRW(cBar, "X Cells", TW_TYPE_UINT32, &st->XCELLS, "min=0 max=500 step = 1 ");
        TwAddVarRW(cBar, "Y Cells", TW_TYPE_UINT32, &st->YCELLS, "min=0 max=500 step = 1 ");
        TwAddButton(cBar, "Change size", ChangeGrid, visualizer, "");
#endif
    }
    
    if (argc == 4)
        factor = atof(argv[3]);
    
    st->autoSolve = true;
    
    st->currentSize[0] = floor(st->currentSize[0]*factor);
    st->solve(false,false);
    glutMainLoop ();
    return 1;
}
